import json
import os
import datetime
from .publisher import GraphPublisher
from .parser import parse_construct_result
from .sparql import SPARQLBuilder


class GraphDriver(object):
    """ Driver for triplestore database """

    def __init__(self, amqp_client, sender_id, repository_name, prefixes_mapping, nodes_namespace_uri, nodes_prefix):
        self._publisher = GraphPublisher(amqp_client, sender_id, repository_name)
        self._sparql_builder = SPARQLBuilder(prefixes_mapping, nodes_namespace_uri, nodes_prefix)
        self._prefixes_mapping = prefixes_mapping
        self._nodes_namespace_uri = nodes_namespace_uri
        self._nodes_prefix = nodes_prefix

    def get_graph_publisher(self):
        return self._publisher

    async def raw_select(self, query: str):
        """ Executes a raw select query
            Is equivalent to get publisher from this object and call its select method
        """
        result = await self._publisher.select(query)
        return result

    async def raw_construct(self, query: str, parse_result=True):
        """ Executes a raw construct query
            if parse_result is True, parse bindings result and convert into properties grouped by entity returned
        """
        bindings = await self._publisher.construct(query)
        if parse_result:
            objects = parse_construct_result(bindings)
        else:
            objects = bindings
        return objects

    async def get_nodes(self, node_type: str):
        """ Get nodes and their properties corresponding to node_type """
        sparql_query = self._sparql_builder.build_construct_objects_query(node_type)
        bindings = await self._publisher.construct(sparql_query)
        objects = parse_construct_result(bindings)
        for obj in objects:
            obj['id'] = self.to_prefixed_uri(obj['id'])
        return objects

    async def get_node(self, node_type: str, object_id: str):
        """ Get node and its properties for node_type and object_id """
        sparql_query = self._sparql_builder.build_construct_object_query(node_type, self.to_absolute_uri(object_id))
        bindings = await self._publisher.construct(sparql_query)
        objects = parse_construct_result(bindings)
        for obj in objects:
            obj['id'] = self.to_prefixed_uri(obj['id'])
        return objects[0]

    def to_absolute_uri(self, uri: str):
        """ Convert an uri into absolute uri """
        if uri.startswith('http://'):
            return uri
        elif self._nodes_prefix in uri:
            return uri.replace(f'{self._nodes_prefix}:', f':{self._nodes_namespace_uri}/')
        else:
            matching_prefix = next((prefix for prefix in self._prefixes_mapping.keys() if prefix in uri), None)
            if matching_prefix:
                return uri.replace(f'{matching_prefix}:', self._prefixes_mapping[matching_prefix])
        # Last case: no prefix at all, we put nodes namespace uri
        return f'{self._nodes_namespace_uri}/{uri}'

    def to_prefixed_uri(self, uri: str):
        """ Convert an uri into prefixed uri """
        if ':' not in uri and '#' not in uri:
            raise Exception(f'Uri {uri} cannot be prefixed with alias as there is no uri prefix')
        if self._nodes_namespace_uri in uri:
            return uri.replace(f'{self._nodes_namespace_uri}/', f'{self._nodes_prefix}:')
        else:
            prefix, ns_uri = next(
                ((prefix, ns_uri) for prefix, ns_uri in self._prefixes_mapping.items() if ns_uri in uri), None
            )
            if ns_uri:
                if '#' in uri:
                    return uri.replace(f'{ns_uri}#', f'{prefix}:')
                else:
                    return uri.replace(f'{ns_uri}/', f'{prefix}:')

        # Last case: no prefix at all, we put nodes prefix
        return f'{self._nodes_prefix}:{uri}'



import os
import json


class TypesRegister(object):
    """ A register for types defined in database
        Types in constructor input must be a list of dicts with keys : { name, node, mongo }
        - name: a name identifier for this type
        - node: node type with prefix used for this type in triplestore
        - mongo: name of mongo collection used for this type
    """

    def __init__(self, types):
        self._types = types
        self.check_types()

    def check_types(self):
        """ Check integrity for types defined """
        prefixes = json.loads(os.environ["NAMESPACE_MAPPING"]).keys()
        for type_def in self._types:
            if "node" not in type_def and "mongo" not in type_def:
                raise Exception(f'Type {type_def["name"]} must be defined at least for "node" or "mongo key"')
            if "node" in type_def:
                if ':' not in type_def["node"]:
                    raise Exception(f'Type {type_def["name"]} has no prefix defined for node type')
                elif type_def["node"].split(':')[0] not in prefixes:
                    raise Exception(f'Type {type_def["name"]} has a node type with a prefix not defined in namespace mapping')

    def exists_type(self, typename: str):
        """ Return if obj_type in argument exists in types registered """
        return self._find_matching_type(typename) is not None

    def resolve_type(self, typename: str):
        """ Returns a tuple (node_type, mongo_collection) for type name in argument """
        type_def = self._find_matching_type(typename)
        graph_type = type_def.get('node', None)
        mongo_collection = type_def.get('mongo', None)
        if mongo_collection:
            mongo_collection = f'{mongo_collection.lower()}s'
        return graph_type, mongo_collection

    def _find_matching_type(self, typename: str):
        """ Find matching type definition for type name in argument """
        for type_def in self._types:
            if type_def['name'] == typename:
                return type_def
        return None



from ..drivers import GraphDriver, MongoDriver
from .types_register import TypesRegister


class Datastore(object):
    """ Datastore for uniformizing query to graph and mongo databases """

    _types_register: TypesRegister
    _graph_driver: GraphDriver
    _mongo_driver: MongoDriver

    def __init__(self, types_register: TypesRegister, graph_driver: GraphDriver = None, mongo_driver: MongoDriver = None):
        self._types_register = types_register
        self._graph_driver = graph_driver
        self._mongo_driver = mongo_driver

    def get_mongo_driver(self):
        return self._mongo_driver

    def get_graph_driver(self):
        return self._graph_driver

    def close_connections(self):
        """ Close driver connections """
        self._mongo_driver.close()

    async def get_objects(self, typename: str):
        """ Return all objects of type defined in argument
            Results are merged objects of triplestore and mongo databases
        """
        if not self._types_register.exists_type(typename):
            raise Exception(f'Unknow type {typename}')

        graph_type, mongo_collection = self._types_register.resolve_type(typename)

        graph_objects = []
        mongo_objects = []
        if self._graph_driver and graph_type:
            graph_objects = await self._graph_driver.get_nodes(graph_type)
        if self._mongo_driver and mongo_collection:
            mongo_objects = self._mongo_driver.find_documents(mongo_collection)

        merged_objects = []
        mongo_objects_id = list(map(lambda x: x['id'], mongo_objects))
        if graph_type:
            for graph_object in graph_objects:
                matching_mongo_object = next((obj for obj in mongo_objects if obj['id'] == graph_object['id']), None) or {}
                merged_objects.append({
                    **graph_object,
                    **matching_mongo_object,
                    "type": typename
                })
        else:
            merged_objects = mongo_objects
        return merged_objects

    async def get_object(self, typename: str, object_id: str):
        """ Return object corresponding to id and type defined in argument
            Result is merge of object found in triplestore and mongo databases if any
        """
        if not self._types_register.exists_type(typename):
            raise Exception(f'Unknow type {typename}')

        graph_type, mongo_collection = self._types_register.resolve_type(typename)
        graph_object = {}
        mongo_object = {}

        if self._graph_driver and graph_type:
            graph_object = await self._graph_driver.get_node(graph_type, object_id)
            graph_object = graph_object or {}
        if self._mongo_driver and mongo_collection:
            mongo_object = self._mongo_driver.find_document(mongo_collection, object_id) or {}

        merged_object = {
            **graph_object,
            **mongo_object,
            "type": typename
        }
        return merged_object



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elz_py_tools',
 'elz_py_tools.datastore',
 'elz_py_tools.drivers',
 'elz_py_tools.drivers.amqp',
 'elz_py_tools.drivers.graph',
 'elz_py_tools.drivers.mongo',
 'elz_py_tools.drivers.toolkit',
 'elz_py_tools.tasks']

package_data = \
{'': ['*']}

install_requires = \
['aio-pika>=6.8.0,<7.0.0', 'pymongo>=3.11.4,<4.0.0']

setup_kwargs = {
    'name': 'elz-py-tools',
    'version': '0.1.0',
    'description': 'Python tools for Elzeard modules',
    'long_description': None,
    'author': 'Hervé',
    'author_email': 'herve.descombe@elzeard.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
